# Copyright (C) 2024  rasmunk
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from gen_vm_image.utils.io import join, remove, exists, makedirs
from gen_vm_image.architecture import load_architecture


class AsyncImageTestContext:
    def __init__(self):
        self.init_done = False

    async def setUp(self):
        if self.init_done:
            return

        # https://cloud.debian.org/images/cloud/bookworm/latest/input_path_image-genericcloud-amd64.qcow2
        # Download the test image for the path input test
        self.test_tmp_directory = join("tests", "tmp")
        self.test_res_directory = join("tests", "res")
        if not exists(self.test_tmp_directory):
            assert makedirs(self.test_tmp_directory)
        self.input_image_path = join(self.test_res_directory, "test.qcow2")
        assert exists(self.input_image_path)

        self.architecture_path = join(
            self.test_res_directory, "advanced_architecture.yml"
        )
        loaded, response = load_architecture(self.architecture_path)
        assert loaded
        assert "architecture" in response

        self.architecture = response["architecture"]
        assert self.architecture is not None
        assert isinstance(self.architecture, dict)

    # Should be used by the non async function tearDownClass to ensure that
    # the following cleanup is done before the class is destroyed
    def tearDown(self):
        assert remove(self.test_tmp_directory, recursive=True)



import argparse
from gen_vm_image.cli.parsers.actions import PositionalArgumentsAction
from gen_vm_image.cli.parsers.single import single_group


def api():
    """Extract the default expected arguments and options from the argparser."""
    parser = argparse.ArgumentParser()
    single_group(parser)
    default_args = []
    default_options = {}

    for action in parser._actions:
        if isinstance(action, argparse._StoreAction):
            # Strip the leading double dash
            striped_key = action.option_strings[1].strip("--")
            key = striped_key.replace("-", "_")
            default_options[key] = action.default
        if isinstance(action, PositionalArgumentsAction):
            default_args.append(action.dest)
    return default_args, default_options



# Copyright (C) 2024  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from gen_vm_image.common.defaults import GENERATED_IMAGE_DIR
from gen_vm_image.cli.cli import add_build_image_cli_arguments
from gen_vm_image.image import generate_image


def corc_initializer_plugin_entrypoint(build_data):
    return generate_image(
        build_data["name"],
        build_data["size"],
        version=build_data.get("version", None),
        input=build_data.get("input", None),
        output_format=build_data.get("format", None),
        output_directory=build_data.get("output_directory", GENERATED_IMAGE_DIR),
        overwrite=build_data.get("overwrite", False),
        verbose=build_data.get("verbose", False),
    )


def gen_vm_image_cli(commands):
    add_build_image_cli_arguments(commands)



# Copyright (C) 2025  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import validators
from gen_vm_image.common.defaults import (
    DEFAULT_BUFFER_SIZE,
    GENERATED_IMAGE_DIR,
    TMP_DIR,
    CONSITENCY_SUPPPORTED_FORMATS,
)
from gen_vm_image.common.codes import (
    PATH_NOT_FOUND_ERROR,
    PATH_NOT_FOUND_ERROR_MSG,
    PATH_CREATE_ERROR,
    PATH_CREATE_ERROR_MSG,
    MISSING_ATTRIBUTE_ERROR,
    MISSING_ATTRIBUTE_ERROR_MSG,
    INVALID_ATTRIBUTE_TYPE_ERROR,
    INVALID_ATTRIBUTE_TYPE_ERROR_MSG,
    CHECKSUM_ERROR,
    RESIZE_ERROR,
    RESIZE_ERROR_MSG,
    CHECK_ERROR,
    CHECK_ERROR_MSG,
    SUCCESS,
    DOWNLOAD_ERROR,
    GETSIZE_ERROR,
    GETSIZE_ERROR_MSG,
)
from gen_vm_image.utils.io import exists, makedirs, hashsum
from gen_vm_image.utils.io import size as get_size
from gen_vm_image.utils.job import run
from gen_vm_image.utils.net import download_file


async def qemu_img_call(action, args, format_output_str=True, verbose=False):
    command = ["qemu-img", action]
    if not verbose:
        command.append("-q")
    command.extend(args)

    result = run(command, format_output_str=format_output_str, capture_output=True)
    if result["returncode"] != "0":
        return False, result["error"]
    return True, result["output"]


async def create_image(path, size, image_format="qcow2", verbose=False):
    args = ["-f", image_format, path, size]
    result, msg = await qemu_img_call("create", args, verbose=verbose)
    if not result:
        return False, msg
    return True, msg


async def convert_image(
    input_path, output_path, input_format="qcow2", output_format="qcow2", verbose=False
):
    args = ["-f", input_format, "-O", output_format, input_path, output_path]
    result, msg = await qemu_img_call("convert", args, verbose=verbose)
    if not result:
        return False, msg
    return True, msg


async def resize_image(
    path, size, image_format="qcow2", resize_args=None, verbose=False
):
    if not resize_args:
        resize_args = []

    result, msg = await qemu_img_call(
        "resize", [*resize_args, "-f", image_format, path, size], verbose=verbose
    )
    if not result:
        return False, msg
    return True, msg


async def info_image(path, info_args=None, format_output_str=True):
    if not info_args:
        info_args = []

    # The info call does not support verbosity/the -q option
    command = ["qemu-img", "info", *info_args, path]
    result = run(command, format_output_str=format_output_str, capture_output=True)
    if result["returncode"] != "0":
        return False, result["error"]
    return True, result["output"]


async def amend_image(path, options, image_format="qcow2", verbose=False):
    args = ["-f", image_format, "-o", options, path]
    result, msg = await qemu_img_call("amend", args, verbose=verbose)
    if not result:
        return False, msg
    return True, msg


async def check_image(path, image_format="qcow2", verbose=False):
    if image_format not in CONSITENCY_SUPPPORTED_FORMATS:
        msg = "format: '{}' not supported for consistency check, only one of: {} is supported".format(
            image_format, CONSITENCY_SUPPPORTED_FORMATS
        )
        return False, msg
    result, msg = await qemu_img_call(
        "check", ["-f", image_format, path], verbose=verbose
    )
    if not result:
        return False, msg
    return True, msg


async def image_size(path):
    # NOTE, this returns a long \n delimitered string with
    # the output from the qemu-img command
    # TODO, improve this to return better structured output
    return await info_image(path)


def expand_byte_magnitude(bytesize):
    # Convert
    expanded_bytesize = None
    if "kib" in bytesize.lower() or "ki" in bytesize.lower():
        expanded_bytesize = int(bytesize.lower().replace("kib", "").replace("ki", ""))
    elif "mib" in bytesize.lower() or "mi" in bytesize.lower():
        expanded_bytesize = (
            int(bytesize.lower().replace("mib", "").replace("mi", "")) * 1024
        )
    elif "mb" in bytesize.lower() or "m" in bytesize.lower():
        expanded_bytesize = (
            int(bytesize.lower().replace("mb", "").replace("m", "")) * 1000
        )
    elif "gib" in bytesize.lower() or "gi" in bytesize.lower():
        expanded_bytesize = (
            int(bytesize.lower().replace("gib", "").replace("gi", "")) * 1024 * 1024
        )
    elif "gb" in bytesize.lower() or "g" in bytesize.lower():
        expanded_bytesize = (
            int(bytesize.lower().replace("gb", "").replace("g", "")) * 1000 * 1000
        )
    elif "tib" in bytesize.lower() or "ti" in bytesize.lower():
        expanded_bytesize = (
            int(bytesize.lower().replace("tib", "").replace("ti", ""))
            * 1024
            * 1024
            * 1024
        )
    elif "tb" in bytesize.lower() or "t" in bytesize.lower():
        expanded_bytesize = (
            int(bytesize.lower().replace("tb", "").replace("t", ""))
            * 1000
            * 1000
            * 1000
        )
    else:
        expanded_bytesize = int(bytesize)
    return expanded_bytesize


async def generate_image(
    name,
    size,
    input=None,
    input_format="qcow2",
    input_checksum_type=None,
    input_checksum=None,
    input_checksum_buffer_size=DEFAULT_BUFFER_SIZE,
    input_checksum_read_bytes=None,
    output_format="qcow2",
    output_directory=GENERATED_IMAGE_DIR,
    overwrite=False,
    verbose=False,
    # Optional version attribute for each image configuration
    version=None,
):
    response = {}
    verbose_outputs = []

    # rename the special input variable to input_
    input_ = input

    if version:
        vm_output_path = os.path.join(
            output_directory,
            "{}-{}.{}".format(name, version, output_format),
        )
    else:
        vm_output_path = os.path.join(
            output_directory, "{}.{}".format(name, output_format)
        )

    # Create the destination directory where the images will be saved
    if not exists(output_directory):
        created = makedirs(output_directory)
        if not created:
            response["msg"] = PATH_CREATE_ERROR_MSG.format(
                output_directory, "Failed to create the images output directory"
            )
            return PATH_CREATE_ERROR, response

    if exists(vm_output_path):
        if verbose:
            verbose_outputs.append(
                "The output image: {} already exists".format(vm_output_path)
            )
        if not overwrite:
            if verbose:
                verbose_outputs.append(
                    "Use the --overwrite flag to overwrite the existing image"
                )
                response["verbose_outputs"] = verbose_outputs
            return SUCCESS, response
        else:
            if verbose:
                verbose_outputs.append(
                    "Overwriting the existing image: {}".format(vm_output_path)
                )

    if input_:
        if not isinstance(input_, str):
            response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                type(input_), input_, "string"
            )
            response["verbose_outputs"] = verbose_outputs
            return INVALID_ATTRIBUTE_TYPE_ERROR, response

        # If a checksum is present, then validate that it is correctly structured
        if input_checksum:
            if not isinstance(input_checksum, str):
                response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                    type(input_checksum),
                    input_checksum,
                    "string",
                )
                response["verbose_outputs"] = verbose_outputs
                return INVALID_ATTRIBUTE_TYPE_ERROR, response

            if not input_checksum_type:
                response["msg"] = MISSING_ATTRIBUTE_ERROR_MSG.format(
                    "input_checksum_type", input_checksum
                )
                response["verbose_outputs"] = verbose_outputs
                return MISSING_ATTRIBUTE_ERROR, response

            if not isinstance(input_checksum_type, str):
                response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                    type(input_checksum_type),
                    input_checksum_type,
                    "string",
                )
                response["verbose_outputs"] = verbose_outputs
                return INVALID_ATTRIBUTE_TYPE_ERROR, response

            if not isinstance(input_checksum_buffer_size, int):
                response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                    type(input_checksum_buffer_size),
                    input_checksum_buffer_size,
                    "int",
                )
                response["verbose_outputs"] = verbose_outputs
                return INVALID_ATTRIBUTE_TYPE_ERROR, response

            if input_checksum_read_bytes and not isinstance(
                input_checksum_read_bytes, int
            ):
                response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                    type(input_checksum_read_bytes),
                    input_checksum_read_bytes,
                    "int",
                )
                response["verbose_outputs"] = verbose_outputs
                return INVALID_ATTRIBUTE_TYPE_ERROR, response

        if validators.url(input_):
            input_url = input_
            # Download the specified url and save it into
            # a tmp directory.
            # First prepare the temporary directory
            # where the downloaded image will be prepared
            if not exists(TMP_DIR):
                created = makedirs(TMP_DIR)
                if not created:
                    response["msg"] = PATH_CREATE_ERROR_MSG.format(
                        TMP_DIR,
                        "Failed to create the temporary download directory",
                    )
                    response["verbose_outputs"] = verbose_outputs
                    return PATH_CREATE_ERROR, response

            input_url_filename = input_url.split("/")[-1]
            input_image_path = os.path.join(TMP_DIR, input_url_filename)
            if not exists(input_image_path):
                if verbose:
                    verbose_outputs.append(
                        "Downloading image from: {}".format(input_url)
                    )
                downloaded, download_response = await download_file(
                    input_url, input_image_path
                )
                if not downloaded:
                    response["msg"] = download_response["msg"]
                    response["verbose_outputs"] = verbose_outputs
                    return DOWNLOAD_ERROR, response
                if verbose:
                    verbose_outputs.append(
                        "Download details: {}".format(download_response)
                    )
        else:
            # If the input_ is a string, then we assume that it is a path to the image
            if not exists(input_):
                response["msg"] = PATH_NOT_FOUND_ERROR_MSG.format(
                    input_, "the defined input_ path to the does not exist"
                )
                response["verbose_outputs"] = verbose_outputs
                return PATH_NOT_FOUND_ERROR, response
            input_image_path = input_

        if not input_format and input_image_path:
            # Try to discover the input_ format since we have
            # only been given a string value
            input_format = input_image_path.split(".")[-1]

        if not isinstance(input_format, str):
            response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                type(input_format), input_format, "string"
            )
            response["verbose_outputs"] = verbose_outputs
            return INVALID_ATTRIBUTE_TYPE_ERROR, response

        if input_checksum:
            calculated_checksum = await hashsum(
                input_image_path,
                algorithm=input_checksum_type,
                buffer_size=input_checksum_buffer_size,
                read_bytes_of_file=input_checksum_read_bytes,
            )
            if not calculated_checksum:
                response["msg"] = (
                    "Failed to calculate the checksum of the downloaded image"
                )
                response["verbose_outputs"] = verbose_outputs
                return CHECKSUM_ERROR, response

            if calculated_checksum != input_checksum:
                response["msg"] = (
                    "The checksum of the downloaded image: {} does not match the expected checksum: {}".format(
                        calculated_checksum, input_checksum
                    )
                )
                response["verbose_outputs"] = verbose_outputs
                return CHECKSUM_ERROR, response
            if verbose:
                verbose_outputs.append(
                    "The calculated checksum: {} matches the defined checksum: {}".format(
                        calculated_checksum, input_checksum
                    )
                )

        converted_result, msg = await convert_image(
            input_image_path,
            vm_output_path,
            input_format=input_format,
            output_format=output_format,
            verbose=verbose,
        )
        if not converted_result:
            response["msg"] = PATH_CREATE_ERROR_MSG.format(input_image_path, msg)
            response["verbose_outputs"] = verbose_outputs
            return PATH_CREATE_ERROR, response

        input_size = get_size(input_image_path)
        if not input_size:
            response["msg"] = GETSIZE_ERROR_MSG.format(input_image_path)
            response["verbose_outputs"] = verbose_outputs
            return GETSIZE_ERROR, response

        expected_resize_size = expand_byte_magnitude(size)
        resize_args = []
        if expected_resize_size < input_size:
            resize_args = ["--shrink"]

        # Resize the vm disk image
        resized_result, resized_msg = await resize_image(
            vm_output_path,
            size,
            image_format=output_format,
            resize_args=resize_args,
            verbose=verbose,
        )
        if not resized_result:
            response["msg"] = RESIZE_ERROR_MSG.format(vm_output_path, resized_msg)
            response["verbose_outputs"] = verbose_outputs
            return RESIZE_ERROR, response
    else:
        # If no input_ is specified, then we assume that we are creating a new disc image
        create_image_result, msg = await create_image(
            vm_output_path,
            size,
            image_format=output_format,
            verbose=verbose,
        )
        if not create_image_result:
            response["msg"] = PATH_CREATE_ERROR_MSG.format(vm_output_path, msg)
            response["verbose_outputs"] = verbose_outputs
            return PATH_CREATE_ERROR, response
        else:
            if verbose:
                verbose_outputs.append(
                    "Generated image at: {}".format(os.path.realpath(vm_output_path))
                )

    # Amend to qcow2 version 3 which is required in RHEL 9 if the output format is
    # qcow2
    # TODO, validate that the image is a rhel based image
    if output_format == "qcow2":
        amend_result, amend_msg = await amend_image(
            vm_output_path, "compat=v3", verbose=verbose
        )
        if not amend_result:
            verbose_outputs.append(
                PATH_CREATE_ERROR_MSG.format(vm_output_path, amend_msg)
            )

    if output_format in CONSITENCY_SUPPPORTED_FORMATS:
        check_result, check_msg = await check_image(
            vm_output_path,
            image_format=output_format,
            verbose=verbose,
        )
        if not check_result:
            response["msg"] = CHECK_ERROR_MSG.format(check_msg)
            response["verbose_outputs"] = verbose_outputs
            return CHECK_ERROR, response
    return SUCCESS, response



# Copyright (C) 2025  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import yaml
from gen_vm_image.common.defaults import GENERATED_IMAGE_DIR, DEFAULT_BUFFER_SIZE
from gen_vm_image.common.codes import (
    PATH_CREATE_ERROR,
    PATH_CREATE_ERROR_MSG,
    PATH_NOT_FOUND_ERROR,
    PATH_NOT_FOUND_ERROR_MSG,
    PATH_LOAD_ERROR,
    PATH_LOAD_ERROR_MSG,
    INVALID_ATTRIBUTE_TYPE_ERROR,
    INVALID_ATTRIBUTE_TYPE_ERROR_MSG,
    MISSING_ATTRIBUTE_ERROR,
    MISSING_ATTRIBUTE_ERROR_MSG,
    SUCCESS,
)
from gen_vm_image.utils.io import exists, load, makedirs
from gen_vm_image.image import generate_image


def load_architecture(architecture_path):
    response = {}
    if not exists(architecture_path):
        response["error_code"] = PATH_NOT_FOUND_ERROR
        response["msg"] = PATH_NOT_FOUND_ERROR_MSG.format(
            architecture_path, "Failed to find the architecture file."
        )
        return False, response

    architecture = load(architecture_path, handler=yaml, Loader=yaml.FullLoader)
    if not architecture:
        response["error_code"] = PATH_LOAD_ERROR
        response["msg"] = PATH_LOAD_ERROR_MSG.format(
            architecture_path, "Failed to load the architecture file."
        )
        return False, response
    response["architecture"] = architecture
    return True, response


def correct_architecture_structure(architecture):
    response = {}
    owner = architecture.get("owner", None)
    if not owner:
        response["error_code"] = MISSING_ATTRIBUTE_ERROR
        response["msg"] = MISSING_ATTRIBUTE_ERROR_MSG.format("owner", architecture)
        return False, response

    images = architecture.get("images", None)
    if not images:
        response["error_code"] = MISSING_ATTRIBUTE_ERROR
        response["msg"] = MISSING_ATTRIBUTE_ERROR_MSG.format("images", architecture)
        return False, response

    required_attributes = ["name", "size"]
    for image_name, image_data in images.items():
        for attribute in required_attributes:
            if attribute not in image_data:
                response["error_code"] = MISSING_ATTRIBUTE_ERROR
                response["msg"] = MISSING_ATTRIBUTE_ERROR_MSG.format(
                    attribute, image_name
                )
                return False, response
    return True, response


def validate_input(input_data):
    response = {}
    if not isinstance(input_data, str) and not isinstance(input_data, dict):
        response["error_code"] = INVALID_ATTRIBUTE_TYPE_ERROR
        response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
            type(input_data),
            input_data,
            "image input must be a string or dictionary",
        )
        return False, response

    if isinstance(input_data, dict):
        if "url" not in input_data and "path" not in input_data:
            response["msg"] = MISSING_ATTRIBUTE_ERROR_MSG.format(
                "'url' or 'path' must be in the architecture input section"
            )

        if "url" in input_data and "path" in input_data:
            response["error_code"] = INVALID_ATTRIBUTE_TYPE_ERROR
            response["msg"] = (
                "Both 'url' and 'path' are defined in the architecture input section. Only one can be defined"
            )
            return False, response

        if "url" in input_data:
            if not isinstance(input_data["url"], str):
                response["error_code"] = INVALID_ATTRIBUTE_TYPE_ERROR
                response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                    type(input_data["url"]), input_data["url"], "string"
                )
                return False, response

        if "path" in input_data:
            if not isinstance(input_data["path"], str):
                response["error_code"] = INVALID_ATTRIBUTE_TYPE_ERROR
                response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                    type(input_data["path"]), input_data["path"], "string"
                )
                return False, response

        if "format" in input_data:
            vm_input_format = input_data["format"]
            if not isinstance(vm_input_format, str):
                response["error_code"] = INVALID_ATTRIBUTE_TYPE_ERROR
                response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                    type(vm_input_format), vm_input_format, "string"
                )
                return False, response

        # If a checksum is present, then validate that it is correctly structured
        if "checksum" in input_data:
            if not isinstance(input_data["checksum"], dict):
                response["error_code"] = INVALID_ATTRIBUTE_TYPE_ERROR
                response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                    type(input_data["checksum"]),
                    input_data["checksum"],
                    "dictionary",
                )
                return False, response

            required_checksum_attributes = ["type", "value"]
            for attr in required_checksum_attributes:
                if attr not in input_data["checksum"]:
                    response["error_code"] = MISSING_ATTRIBUTE_ERROR
                    response["msg"] = MISSING_ATTRIBUTE_ERROR_MSG.format(
                        attr, input_data["checksum"]
                    )
                    return False, response
                if not isinstance(input_data["checksum"][attr], str):
                    response["error_code"] = INVALID_ATTRIBUTE_TYPE_ERROR
                    response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                        type(input_data["checksum"][attr]),
                        input_data["checksum"][attr],
                        "string",
                    )
                    return False, response
    return True, response


def prepare_input_kwargs(input_data):
    input_kwargs = {}

    checksum = input_data.get("checksum", None)
    if checksum:
        input_kwargs["input_checksum_type"] = checksum.get("type", None)
        input_kwargs["input_checksum"] = checksum.get("value", None)
        input_kwargs["input_checksum_buffer_size"] = checksum.get(
            "buffer_size", DEFAULT_BUFFER_SIZE
        )
        input_kwargs["input_checksum_read_bytes"] = checksum.get("read_bytes", None)

    if "path" in input_data:
        input_kwargs["input"] = input_data.get("path", None)
    if "url" in input_data:
        input_kwargs["input"] = input_data.get("url", None)
    if "format" in input_data:
        input_kwargs["input_format"] = input_data.get("format", None)
    return input_kwargs


async def build_architecture(
    architecture_path,
    output_directory=GENERATED_IMAGE_DIR,
    overwrite=False,
    verbose=False,
):
    response = {"verbose_outputs": []}
    # Load the architecture file
    architecture_loaded, architecture_response = load_architecture(architecture_path)
    if not architecture_loaded:
        response["msg"] = architecture_response["msg"]
        return architecture_response["error_code"], response

    architecture = architecture_response["architecture"]
    correct_architecture, correct_response = correct_architecture_structure(
        architecture
    )
    if not correct_architecture:
        response["msg"] = correct_response["msg"]
        return correct_response["error_code"], response

    # Create the destination directory where the images will be saved
    if not exists(output_directory):
        created = makedirs(output_directory)
        if not created:
            response["msg"] = PATH_CREATE_ERROR_MSG.format(
                output_directory, "Failed to create the images output directory"
            )
            return PATH_CREATE_ERROR, response

    if "images" not in architecture:
        response["msg"] = MISSING_ATTRIBUTE_ERROR_MSG.format("images", architecture)
        return MISSING_ATTRIBUTE_ERROR, response

    images = architecture.get("images")
    if not isinstance(images, dict):
        response["msg"] = INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
            type(images), images, dict
        )
        return INVALID_ATTRIBUTE_TYPE_ERROR, response

    # Generate the image configuration
    for _, build_data in images.items():
        generate_image_kwargs = {}

        input_ = build_data.get("input", None)
        if input_:
            correct_input, correct_input_response = validate_input(input_)
            if not correct_input:
                return (
                    correct_input_response["error_code"],
                    correct_input_response["msg"],
                )

            if isinstance(input_, dict):
                generate_image_kwargs.update(**prepare_input_kwargs(input_))

        generate_image_kwargs["output_directory"] = output_directory
        generate_image_kwargs["output_format"] = build_data.get("format", "qcow2")
        generate_image_kwargs["version"] = build_data.get("version", None)

        build_return_code, build_response = await generate_image(
            build_data["name"],
            build_data["size"],
            **generate_image_kwargs,
            verbose=verbose,
        )

        if build_return_code != SUCCESS:
            response["verbose_outputs"] = build_response.get("verbose_outputs", [])
            response["msg"] = build_response.get("msg", "")
            return build_return_code, response

    response["msg"] = "Successfully built the images in: {}".format(
        os.path.realpath(output_directory)
    )
    return SUCCESS, response



# Copyright (C) 2025  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import unittest
import os
import random
from gen_vm_image.common.defaults import MULTIPLE
from gen_vm_image.common.codes import SUCCESS
from gen_vm_image.utils.io import exists, makedirs, join, remove
from gen_vm_image.cli.cli import main

BASE_IMAGES_DIRECTORY = os.path.realpath(join("tests", "tmp", "images"))

BASIC_ARCHITECTURE_FILE = "basic_architecture.yml"
BASIC_ARCHITECTURE_PATH = os.path.realpath(
    join("tests", "res", BASIC_ARCHITECTURE_FILE)
)

ADVANCED_ARCHITECTURE_FILE = "advanced_architecture.yml"
ADVANCED_ARCHITECTURE_PATH = os.path.realpath(
    join("tests", "res", ADVANCED_ARCHITECTURE_FILE)
)


class TestCLIMultipleImage(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.seed = str(random.random())[2:10]
        cls.images_dir = join(BASE_IMAGES_DIRECTORY, cls.seed)
        cls.res_dir = os.path.realpath(join("tests", "res"))
        if not exists(cls.images_dir):
            assert makedirs(cls.images_dir)

    @classmethod
    def tearDownClass(cls):
        if exists(cls.images_dir):
            assert remove(cls.images_dir, recursive=True)

    def test_cli_help(self):
        return_code = None
        try:
            return_code = main([MULTIPLE, "--help"])
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)

    def test_cli_basic_multiple_images(self):
        return_code = None
        try:
            self.assertTrue(exists(BASIC_ARCHITECTURE_PATH))
            return_code = main(
                [
                    MULTIPLE,
                    BASIC_ARCHITECTURE_PATH,
                    "--output-directory",
                    self.images_dir,
                ]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)
        expected_ouput_image_name = "test-image-9.4.qcow2"
        expected_output_image = join(self.images_dir, expected_ouput_image_name)
        self.assertTrue(exists(expected_output_image))
        # TODO verify the size is correct

    def test_cli_advanced_multiple_images(self):
        return_code = None
        try:
            self.assertTrue(exists(ADVANCED_ARCHITECTURE_PATH))
            return_code = main(
                [
                    MULTIPLE,
                    ADVANCED_ARCHITECTURE_PATH,
                    "--output-directory",
                    self.images_dir,
                ]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)
        expected_images = [
            "test-image-1-9.4.qcow2",
            "test-image-2-9.4.raw",
            "test-image-3-8.4.qcow2",
            "test-image-4-8.4.raw",
            "test-image-5-9.4.raw",
            "input-path-image-12.qcow2",
            "convert_input_image_format-12.raw",
            "non-version-image.raw",
            "image-with-output-path-9.4.qcow2",
        ]
        for image in expected_images:
            expected_output_image = join(self.images_dir, image)
            self.assertTrue(exists(expected_output_image))



# Copyright (C) 2025  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import argparse


class PositionalArgumentsAction(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if not hasattr(namespace, "positional_arguments") or not getattr(
            namespace, "positional_arguments"
        ):
            setattr(namespace, "positional_arguments", [values])
        else:
            getattr(namespace, "positional_arguments").append(values)



# Copyright (C) 2025  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from gen_vm_image.common.defaults import (
    SINGLE,
    GENERATED_IMAGE_DIR,
    DEFAULT_BUFFER_SIZE,
)
from gen_vm_image.cli.parsers.actions import PositionalArgumentsAction


def single_group(parser):
    generate_single_group = parser.add_argument_group(
        title="Generate a single Virtual Machine Image"
    )

    generate_single_group.add_argument(
        "name",
        action=PositionalArgumentsAction,
        type=str,
        help="The name of the image that will be generated.",
    )
    generate_single_group.add_argument(
        "size",
        action=PositionalArgumentsAction,
        type=str,
        help="The size of the image that will be generated.",
    )
    generate_single_group.add_argument(
        "-i",
        "--input",
        dest="{}_input".format(SINGLE),
        default=None,
        help="The path or url to the input image that the generated image should be based on.",
    )
    generate_single_group.add_argument(
        "-if",
        "--input-format",
        dest="{}_input_format".format(SINGLE),
        default=None,
        help="The format of the input image. Will dynamically try to determine the format if not provided.",
    )
    generate_single_group.add_argument(
        "-ict",
        "--input-checksum-type",
        dest="{}_input_checksum_type".format(SINGLE),
        default=None,
        help="The checksum type that should be used to validate the input image if set.",
    )
    generate_single_group.add_argument(
        "-ic",
        "--input-checksum",
        dest="{}_input_checksum".format(SINGLE),
        default=None,
        help="The checksum that should be used to validate the input image if set.",
    )
    generate_single_group.add_argument(
        "-icbs",
        "--input-checksum-buffer-size",
        dest="{}_input_checksum_buffer_size".format(SINGLE),
        default=DEFAULT_BUFFER_SIZE,
        help="The buffer size that is used to read the input image when calculating the checksum value.",
    )
    generate_single_group.add_argument(
        "-icrb",
        "--input-checksum-read-bytes",
        dest="{}_input_checksum_read_bytes".format(SINGLE),
        default=None,
        type=int,
        help="The amount of bytes that should be read from the input image to be used to calculate the expected checksum value.",
    )
    generate_single_group.add_argument(
        "-od",
        "--output-directory",
        dest="{}_output_directory".format(SINGLE),
        default=GENERATED_IMAGE_DIR,
        help="The path to the output directory where the image will be saved.",
    )
    generate_single_group.add_argument(
        "-of",
        "--output-format",
        dest="{}_output_format".format(SINGLE),
        default="qcow2",
        help="The format of the output image.",
    )
    generate_single_group.add_argument(
        "-V",
        "--version",
        dest="{}_version".format(SINGLE),
        help="The version of the image that is generated.",
    )
    generate_single_group.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        dest="{}_verbose".format(SINGLE),
        default=False,
        help="Print verbose output.",
    )



# Copyright (C) 2024  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import argparse
import asyncio
import json
import inspect
import sys
import os
from gen_vm_image._version import __version__
from gen_vm_image.common.defaults import PACKAGE_NAME, GEN_VM_IMAGE_CLI_STRUCTURE
from gen_vm_image.common.codes import (
    SUCCESS,
    JSON_DUMP_ERROR,
    JSON_DUMP_ERROR_MSG,
)
from gen_vm_image.cli.common import error_print, to_str
from gen_vm_image.cli.helpers import extract_arguments, strip_argument_group_prefix


current_dir = os.path.realpath(os.path.dirname(__file__))
parent_dir = os.path.dirname(current_dir)


def import_from_module(module_path, module_name, func_name):
    module = __import__(module_path, fromlist=[module_name])
    return getattr(module, func_name)


def add_cli_operations(
    parser,
    operation,
    module_cli_input_group_prefix="gen_vm_image.cli.input_groups",
    module_operation_prefix="gen_vm_image.cli.operations",
):
    operation_input_groups_func = import_from_module(
        "{}.{}".format(module_cli_input_group_prefix, operation),
        "{}".format(operation),
        "{}_groups".format(operation),
    )

    provider_groups = []
    argument_groups = []
    input_groups = operation_input_groups_func(parser)
    if not input_groups:
        raise RuntimeError(
            "No input groups were returned by the input group function: {}".format(
                operation_input_groups_func.func_name
            )
        )

    argument_groups = input_groups
    parser.set_defaults(
        func=cli_exec,
        module_path="{}.{}.build".format(module_operation_prefix, operation),
        module_name="build",
        func_name="{}_operation".format(operation),
        provider_groups=provider_groups,
        argument_groups=argument_groups,
    )


def cli_exec(arguments):
    # Actions determines which function to execute
    module_path = arguments.pop("module_path")
    module_name = arguments.pop("module_name")
    func_name = arguments.pop("func_name")

    if "positional_arguments" in arguments:
        positional_arguments = arguments.pop("positional_arguments")
    else:
        positional_arguments = []

    if "argument_groups" in arguments:
        argument_groups = arguments.pop("argument_groups")
    else:
        argument_groups = []

    func = import_from_module(module_path, module_name, func_name)
    if not func:
        return False, {}

    action_kwargs, _ = extract_arguments(arguments, argument_groups)
    action_kwargs = strip_argument_group_prefix(action_kwargs, argument_groups)

    action_args = positional_arguments
    if inspect.iscoroutinefunction(func):
        return asyncio.run(func(*action_args, **action_kwargs))
    return func(*action_args, **action_kwargs)


def add_base_cli_operations(parser):
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=__version__,
        help="Print the version of the program",
    )


def add_build_image_cli_arguments(
    commands, module_cli_prefix="gen_vm_image.cli.operations"
):
    for command in GEN_VM_IMAGE_CLI_STRUCTURE:
        command_parser = commands.add_parser(command)
        add_cli_operations(command_parser, command)


def main(args):
    parser = argparse.ArgumentParser(
        prog=PACKAGE_NAME, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    # Add the basic CLI functions
    add_base_cli_operations(parser)
    # Add the build image CLI arguments
    commands = parser.add_subparsers(title="COMMAND")
    add_build_image_cli_arguments(commands)

    parsed_args = parser.parse_args(args)
    # Convert to a dictionary
    arguments = vars(parsed_args)

    if "func" in arguments:
        func = arguments.pop("func")
        return_code, result_dict = func(arguments)

        response = {}
        if return_code == SUCCESS:
            response["status"] = "success"
        else:
            response["status"] = "failed"
        response["verbose_messages"] = result_dict.get("verbose_outputs", [])
        response["response"] = result_dict.get("msg", "")
        response["return_code"] = return_code

        try:
            output = json.dumps(response, indent=4, sort_keys=True, default=to_str)
        except Exception as err:
            error_print(JSON_DUMP_ERROR_MSG.format(err))
            return JSON_DUMP_ERROR
        if return_code == SUCCESS:
            print(output)
        else:
            error_print(output)
        return return_code
    return SUCCESS


def cli():
    return main(sys.argv[1:])


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright (C) 2025  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from gen_vm_image.common.defaults import MULTIPLE, GENERATED_IMAGE_DIR
from gen_vm_image.cli.parsers.actions import PositionalArgumentsAction


def multiple_group(parser):
    generate_multiple_group = parser.add_argument_group(
        title="Generate multiple Virtual Machine Images"
    )

    generate_multiple_group.add_argument(
        "architecture_path",
        action=PositionalArgumentsAction,
        help="The path to the architecture file that defines the images that should be generated.",
    )
    generate_multiple_group.add_argument(
        "-od",
        "--output-directory",
        dest="{}_output_directory".format(MULTIPLE),
        default=GENERATED_IMAGE_DIR,
        help="The path to the output directory where the images will be saved.",
    )
    generate_multiple_group.add_argument(
        "--overwrite",
        dest="{}_overwrite".format(MULTIPLE),
        action="store_true",
        default=False,
        help="Whether the tool should overwrite existing image disks.",
    )
    generate_multiple_group.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        dest="{}_verbose".format(MULTIPLE),
        default=False,
        help="Print verbose output.",
    )



# Copyright (C) 2025  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


def strip_argument_prefix(arguments, prefix=""):
    return {k.replace(prefix, ""): v for k, v in arguments.items()}


def get_arguments(arguments, startswith=""):
    return {k: v for k, v in arguments.items() if k.startswith(startswith)}


def extract_arguments(arguments, argument_groups):
    found_kwargs, remaining_kwargs = {}, {}
    for argument_group in argument_groups:
        group_args = get_arguments(arguments, argument_group.lower())
        found_kwargs.update(group_args)
    remaining_kwargs = {
        k: v for k, v in arguments.items() if k not in found_kwargs and v
    }
    return found_kwargs, remaining_kwargs


def strip_argument_group_prefix(arguments, argument_groups):
    args = {}
    for argument_group in argument_groups:
        group_arguments = get_arguments(arguments, argument_group.lower())
        args.update(
            strip_argument_prefix(group_arguments, argument_group.lower() + "_")
        )
    return args



# Copyright (C) 2025  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import unittest
import os
import random
from gen_vm_image.common.codes import SUCCESS, CHECKSUM_ERROR
from gen_vm_image.common.defaults import SINGLE
from gen_vm_image.utils.io import makedirs, exists, remove, join
from gen_vm_image.cli.cli import main


TEST_RES_DIR = os.path.realpath(join("tests", "res"))
TEST_IMAGE_NAME = "test.qcow2"
TEST_IMAGE_PATH = os.path.realpath(join(TEST_RES_DIR, TEST_IMAGE_NAME))


class TestCLISingleImage(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.seed = str(random.random())[2:10]
        cls.images_dir = os.path.realpath(join("tests", "tmp", "images", cls.seed))
        if not exists(cls.images_dir):
            assert makedirs(cls.images_dir)

    @classmethod
    def tearDownClass(cls):
        if exists(cls.images_dir):
            assert remove(cls.images_dir, recursive=True)

    def test_cli_help(self):
        return_code = None
        try:
            return_code = main([SINGLE, "--help"])
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)

    def test_cli_basic_single_image(self):
        return_code = None
        try:
            name = "test-cli-single-image-{}".format(self.seed)
            size = "5G"
            return_code = main(
                [SINGLE, name, size, "--output-directory", self.images_dir]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)
        output_path = join(self.images_dir, "{}.qcow2".format(name))
        self.assertTrue(exists(output_path))

    def test_cli_single_image_with_output_format_raw(self):
        return_code = None
        try:
            name = "test-cli-single-image-output-format-{}".format(self.seed)
            size = "5G"
            output_format = "raw"
            return_code = main(
                [
                    SINGLE,
                    name,
                    size,
                    "--output-directory",
                    self.images_dir,
                    "--output-format",
                    output_format,
                ]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)

        output_path = join(self.images_dir, "{}.{}".format(name, output_format))
        self.assertTrue(exists(output_path))

    def test_cli_single_image_with_output_format_qcow2(self):
        return_code = None
        try:
            name = "test-cli-single-image-output-format-{}".format(self.seed)
            size = "5G"
            output_format = "qcow2"
            return_code = main(
                [
                    SINGLE,
                    name,
                    size,
                    "--output-directory",
                    self.images_dir,
                    "--output-format",
                    output_format,
                ]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)

        output_path = join(self.images_dir, "{}.{}".format(name, output_format))
        self.assertTrue(exists(output_path))

    def test_cli_single_image_with_version(self):
        return_code = None
        try:
            name = "test-cli-single-image-version-{}".format(self.seed)
            size = "5G"
            version = "1.0"
            return_code = main(
                [
                    SINGLE,
                    name,
                    size,
                    "--output-directory",
                    self.images_dir,
                    "--version",
                    version,
                ]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)

        output_path = join(self.images_dir, "{}-{}.qcow2".format(name, version))
        self.assertTrue(exists(output_path))

    def test_cli_single_image_with_input_path(self):
        return_code = None
        try:
            name = "test-cli-single-image-input-{}".format(self.seed)
            size = "5G"
            return_code = main(
                [
                    SINGLE,
                    name,
                    size,
                    "--output-directory",
                    self.images_dir,
                    "--input",
                    TEST_IMAGE_PATH,
                ]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)

        output_path = join(self.images_dir, "{}.qcow2".format(name))
        self.assertTrue(exists(output_path))

    def test_cli_single_image_with_input_path_and_raw_output_format(self):
        return_code = None
        try:
            name = "test-cli-single-image-input-{}".format(self.seed)
            size = "5G"
            output_format = "raw"
            return_code = main(
                [
                    SINGLE,
                    name,
                    size,
                    "--output-directory",
                    self.images_dir,
                    "--input",
                    TEST_IMAGE_PATH,
                    "--output-format",
                    output_format,
                ]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)

        output_path = join(self.images_dir, "{}.{}".format(name, output_format))
        self.assertTrue(exists(output_path))

    def test_cli_single_image_with_input_url(self):
        return_code = None
        try:
            name = "test-cli-single-image-input-{}".format(self.seed)
            size = "5G"
            image_url = "https://download.rockylinux.org/pub/rocky/9/images/x86_64/Rocky-9-GenericCloud-Base.latest.x86_64.qcow2"
            return_code = main(
                [
                    SINGLE,
                    name,
                    size,
                    "--output-directory",
                    self.images_dir,
                    "--input",
                    image_url,
                ]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)

        output_path = join(self.images_dir, "{}.qcow2".format(name))
        self.assertTrue(exists(output_path))

    def test_cli_single_image_with_checksum(self):
        expected_checksum = "db12db337b0d1c82ab5b174c21241e0be24ee44a79e0d310aa05e3652e301a22bf604fd86c4cf9e4489bd512cab275feb8d8326847ca04b90e0903c72f39cb64"
        return_code = None
        try:
            name = "test-cli-single-image-checksum-{}".format(self.seed)
            size = "10G"
            return_code = main(
                [
                    SINGLE,
                    name,
                    size,
                    "--output-directory",
                    self.images_dir,
                    "--input",
                    TEST_IMAGE_PATH,
                    "--input-checksum-type",
                    "sha512",
                    "--input-checksum",
                    expected_checksum,
                ]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)
        output_path = join(self.images_dir, "{}.qcow2".format(name))
        self.assertTrue(exists(output_path))

    def test_cli_single_image_incorrect_checksum(self):
        expected_checksum = "12931j3901j39183189231230912309-01230-0-askdasidma"
        return_code = None
        try:
            name = "test-cli-single-image-checksum-{}".format(self.seed)
            size = "10G"
            return_code = main(
                [
                    SINGLE,
                    name,
                    size,
                    "--output-directory",
                    self.images_dir,
                    "--input",
                    TEST_IMAGE_PATH,
                    "--input-checksum-type",
                    "sha512",
                    "--input-checksum",
                    expected_checksum,
                ]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, CHECKSUM_ERROR)
        output_path = join(self.images_dir, "{}.qcow2".format(name))
        self.assertFalse(exists(output_path))

    def test_cli_single_image_with_checksum_and_partial_bytes(self):
        expected_checksum = "1d7e67951f6b693b8f0ec66c7ad27743edb9088bee19b60cdd564abe0f3d1f36664299aa03ad4903e60fccdd60faf71b4eacb498c8c33da06bd5f6f33f7852ce"
        return_code = None
        try:
            name = "test-cli-single-image-checksum-partial-bytes-{}".format(self.seed)
            size = "5G"
            read_bytes_of_file = "1000"
            return_code = main(
                [
                    SINGLE,
                    name,
                    size,
                    "--output-directory",
                    self.images_dir,
                    "--input",
                    TEST_IMAGE_PATH,
                    "--input-checksum-type",
                    "sha512",
                    "--input-checksum",
                    expected_checksum,
                    "--input-checksum-read-bytes",
                    read_bytes_of_file,
                ]
            )
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)
        output_path = join(self.images_dir, "{}.qcow2".format(name))
        self.assertTrue(exists(output_path))
